import datetime
import os
import re

base_dir = os.path.expanduser("~/.azx")


class Store:
    def __init__(self):
        self.started_at = _now_str()
        self.ended_at = self.started_at
        self.progress = 0
        self.conversation = []

    def tool(self, id, name, args, ret):
        self._add_tool_to_last_assistant_msg(id, name, args)
        self.ended_at = _now_str()
        self.conversation.append(
            {
                "role": "tool",
                "tool_call_id": id,
                "name": name,
                "content": ret,
            }
        )

        os.makedirs(self._loc(), exist_ok=True)
        with open(self._log_path("tool"), "w") as f:
            f.write("\n".join([id, name, args, ret]))

    def log(self, role: str, msg: str):
        self.ended_at = _now_str()
        self.conversation.append({"role": role, "content": msg})

        os.makedirs(self._loc(), exist_ok=True)
        with open(self._log_path(role), "w") as f:
            f.write(msg)

    def summary(self, sum: str):
        with open(os.path.join(self._loc(), f"{self.ended_at}.sum.md"), "w") as f:
            f.write(sum)

    def sum_or_quest(self):
        def last_summary():
            files = [
                f
                for f in os.listdir(self._loc())
                if os.path.isfile(os.path.join(self._loc(), f)) and f.endswith("sum.md")
            ]

            if not files:
                return ""

            files.sort()

            with open(os.path.join(self._loc(), files[-1]), "r") as f:
                return f.read().strip()

        def first_question():
            for speak in self.conversation:
                if speak["role"] == "user":
                    return speak["content"]

        return last_summary() or first_question() or "nothing"

    def resume(self, started_at: str):
        self.started_at = started_at

        dir_path = self._loc()
        if not os.path.exists(dir_path):
            return

        files = [
            f
            for f in os.listdir(dir_path)
            if os.path.isfile(os.path.join(dir_path, f))
            and (
                f.endswith("user.md")
                or f.endswith("system.md")
                or f.endswith("assistant.md")
                or f.endswith("tool.md")
            )
        ]
        files.sort()

        if not files:
            return

        self.ended_at = os.path.basename(files[-1]).split(".")[0]
        self.progress = len(files)
        self.conversation.clear()

        for filename in files:
            file_path = os.path.join(dir_path, filename)
            try:
                with open(file_path, "r") as f:
                    segments = filename.split(".")
                    role = segments[-2]
                    if role == "tool":
                        fn_id = f.readline().strip()
                        fn_name = f.readline().strip()
                        fn_args = f.readline().strip()
                        fn_ret = f.read().strip()
                        self._add_tool_to_last_assistant_msg(fn_id, fn_name, fn_args)
                        self.conversation.append(
                            {
                                "role": role,
                                "tool_call_id": fn_id,
                                "name": fn_name,
                                "content": fn_ret,
                            }
                        )
                    else:
                        self.conversation.append(
                            {"role": role, "content": f.read().strip()}
                        )
            except Exception:
                continue

    def __str__(self):
        return f"**{self.started_at}** ~ **{self.ended_at}**: {self.sum_or_quest()}"

    def _loc(self) -> str:
        return os.path.join(base_dir, self.started_at)

    def _log_path(self, role) -> str:
        file_name = f"{self.ended_at}.{self.progress}.{role}.md"
        full_path = os.path.join(self._loc(), file_name)
        self.progress += 1
        return full_path

    def _add_tool_to_last_assistant_msg(self, id, name, args):
        last_msg = next(
            (msg for msg in reversed(self.conversation) if msg["role"] == "assistant"),
            None,
        )

        if "tool_calls" not in last_msg:
            last_msg["tool_calls"] = []

        last_msg["tool_calls"].append(
            {
                "id": id,
                "type": "function",
                "function": {"name": name, "arguments": args},
            }
        )


def _now_str() -> str:
    return datetime.datetime.now().strftime("%Y_%m%d_%H%M%S")


def history() -> str:
    if not os.path.exists(base_dir):
        return "No history found."

    def resume(started_at: str) -> Store:
        store = Store()
        store.resume(started_at)
        return store

    stores = [
        resume(item)
        for item in os.listdir(base_dir)
        if os.path.isdir(os.path.join(base_dir, item))
        and re.match(r"^\d{4}_\d{4}_\d{6}$", item)
    ]

    stores.sort(key=lambda s: s.ended_at, reverse=True)

    items = [f"{i + 1}. {store}" for i, store in enumerate(stores)]

    return "\n".join(items) if items else "No history found."



import asyncio
import json
import re
import traceback

from . import prompt
from . import arguments
from .agents import Client
from .configure import Configure
from .renderer import (
    render_error,
    render_md_full,
    render_md_stream,
    render_tool_call,
    render_user_input,
)
from .storage import Store, history
from .tools import Calls, Tools

config = Configure()
args = arguments.parse()


class Chat:
    def __init__(self):
        self.model = config.default_chat_model()
        self.tools = Tools()
        self.session = prompt.session()
        self.store = None

    async def run(self):
        await self._new_client()

        while True:
            try:
                user_input = await self.session.prompt_async()

                # handle command
                user_cmd = user_input.strip().lower()
                if user_cmd in ("/q", "/quit"):
                    break

                if await self._other_command(user_cmd):
                    continue

                # handle chat
                if self.store is None:
                    self.store = Store()
                    self.store.log("system", config.default_chat_prompt())

                self.store.log("user", user_input)

                while True:
                    content, tool_calls = self.client.stream_response(
                        self.store.conversation
                    )
                    whole_output = render_md_stream(content)
                    self.store.log("assistant", whole_output)
                    calls = Calls(tool_calls)
                    for call in calls:
                        render_tool_call(f"{call.fn}({call.params_str()})")
                        self.store.tool(
                            call.id,
                            call.fn,
                            call.params_str(),
                            await self.tools.execute(call),
                        )
                    if not len(calls):
                        break

            except Exception as e:
                render_error(f"Error: {e}\n{traceback.format_exc()}")

    async def _new_client(self):
        self.client = Client(**(self.model | {"tools": await self.tools.specs()}))

    async def _other_command(self, user_cmd):
        if match := re.match(r"^(?:/c|/client)$", user_cmd):
            render_md_full(f"clients:\n{config.models()}")
            return True

        if match := re.match(r"^(?:/c|/client) (.+)$", user_cmd):
            name = match.group(1)
            client2_cfg = config.find_model(name)
            if client2_cfg:
                self.model = client2_cfg
                await self._new_client()
                print(f"Switched to client: {client2_cfg['name']}")
            else:
                print(f"Client '{name}' not found in config")
            return True

        if user_cmd in ("/n", "/new"):
            self.store = None
            return True

        if match := re.match(r"^(?:/r|/resume)$", user_cmd):
            render_md_full(f"history:\n{history()}")
            return True

        if match := re.match(r"^(?:/r|/resume) (.+)$", user_cmd):
            started_at = (
                history().split("\n")[int(match.group(1)) - 1].split(" ")[1].strip("*")
            )
            self.store = Store()
            self.store.resume(started_at)
            for msg in self.store.conversation:
                if msg["role"] == "user":
                    render_user_input(msg["content"])
                elif msg["role"] == "assistant":
                    render_md_stream(msg["content"])
                    for fn in msg.get("tool_calls", []):
                        render_tool_call(
                            f"{fn['function']['name']}({fn['function']['arguments']})"
                        )
            return True

        if user_cmd in ("/s", "/sum", "/summary"):
            talk = self.store.conversation.copy()
            talk.append(
                {
                    "role": "user",
                    "content": "Summarize all talk above briefly, use single language, which is the primary language involved, with words or phrases, in one line. Your answer could contain verb/object/attribute/adverbial/complement, but no subject. Just give me the answer, no thought is need",
                }
            )
            chunked_sum, _ = self.client.stream_response(talk)
            sum = "".join(list(chunked_sum))
            self.store.summary(sum)
            render_md_stream([sum])
            return True

        if match := re.match(r"^(?:/t|/tool)\+ (.+)$", user_cmd):
            await self.tools.add_mcp(match.group(1))
            await self._new_client()
            return True

        if match := re.match(r"^(?:/t|/tool)\- (.+)$", user_cmd):
            await self.tools.del_mcp(match.group(1))
            await self._new_client()
            return True

        if user_cmd in ("/?", "/help") or re.match(r"^/[a-zA-Z0-9]+$", user_cmd):
            manual = "\n".join(
                [
                    f"- {cmd}"
                    for cmd in [
                        "/? /help",
                        "/c /client",
                        "/n /new",
                        "/r /resume",
                        "/s /sum /summary",
                        "/q /quit",
                    ]
                ],
            )
            render_md_full(f"commands:\n{manual}")
            return True

        return False


def ocr():
    model_cfg = (
        config.find_model(args.model) if args.model else config.default_cli_ocr_model()
    )
    client = Client(**model_cfg)
    result = None

    try:
        result = client.ocr(args.files[0])
    except Exception as e:
        print(f"Fail to OCR: {e}")
        traceback.print_exc()

    try:
        js = json.loads(result)
        print(f"{js['abstract']}\n\n{js['full']}")
    except Exception as e:
        print(f"Fail to parse as json: {e}\n\n{result}")
        traceback.print_exc()


def main():
    if args.models:
        return render_md_full(f"clients:\n{config.models()}")

    if args.ocr:
        if not args.files:
            print("Error: --ocr-md or --ocr-json requires a file path argument")
            return
        return ocr()

    asyncio.run(Chat().run())


if __name__ == "__main__":
    main()



import argparse


def parse():
    parser = argparse.ArgumentParser()

    parser.add_argument("--ocr", action="store_true", help="OCR.")
    parser.add_argument("--models", action="store_true", help="List models")
    parser.add_argument("--model", type=str, help="Select a model", default=None)
    parser.add_argument("files", nargs="*", help="One or more file paths.")

    return parser.parse_args()



from markdown_it import MarkdownIt
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.text import Text
from rich.theme import Theme

console = Console(theme=Theme({"markdown.hr": "medium_purple4"}))


def render_error(string):
    text = Text(string)
    text.stylize("red")
    console.print(text)


def render_user_input(string):
    text = Text(f">>> {string}")
    text.stylize("green")
    console.print(text)


def render_tool_call(string):
    text = Text(f"{string}...")
    text.stylize("bright_black")
    console.print(text)


def render_md_full(string):
    console.print(Markdown(string))


def render_md_stream(strings) -> str:
    whole_string = ""
    current_block = ""

    def flatten_strings():
        for string in strings:
            for char in string:
                yield char

    def block_recognized():
        nonlocal current_block
        md = MarkdownIt("js-default")
        tokens_len = 0
        for char in flatten_strings():
            current_block += char
            tokens = md.parse(current_block)
            new_block = (
                tokens_len != len(tokens)
                and current_block[-3:][:2] == "\n\n"
                and char not in "-*123456789|"
            )
            tokens_len = len(tokens)
            yield (char, new_block)

    def new_live() -> Live:
        live = Live(console=console, refresh_per_second=25)
        live.start()
        return live

    live = new_live()

    for char, new_block in block_recognized():
        if new_block:
            current_block = char
            live.stop()
            console.print("")
            live = new_live()

        whole_string += char
        live.update(Markdown(current_block))

    live.stop()

    return whole_string



from pathlib import Path

import yaml

sys_prompt = """
You are an assistant integrated with a chat loop that can call external functions to enhance responses.

When a function is called, it returns a JSON object with the following structure:

{
  "status": "success | error | partial",
  "data": "xxx",// Primary result or response data. It will be None unless status success 
  "err": "xxx", // Human-readable message for context or errors for. It will be None unless status error
  "proceed": "xxx" // What to do next. It will be None unless status partial
}
""".strip()


class Configure:
    def __init__(self):
        config_path = Path.home() / ".azx" / "config.yaml"
        with open(config_path) as f:
            self.config = yaml.safe_load(f)

    def models(self) -> str:
        return "\n".join(
            [f"{i + 1}. {k['name']}" for i, k in enumerate(self.config["keys"])]
        )

    def find_model(self, name_or_num) -> dict:
        for i, k in enumerate(self.config["keys"]):
            if k["name"] == name_or_num or str(i + 1) == name_or_num:
                return k
        return None

    def default_chat_model(self) -> dict:
        return self.config["keys"][0]

    def default_chat_prompt(self) -> dict:
        custom_prompt = self.config.get("prompt", None)
        if custom_prompt is None:
            return sys_prompt
        return f"{sys_prompt}\n\n{custom_prompt.strip()}"

    def default_cli_ocr_model(self) -> dict:
        for k in self.config["keys"]:
            if self.config["cli_ocr"] == k["name"]:
                return k
        return None



import base64
import os
from itertools import tee
from pathlib import Path

from openai import OpenAI


_mime_types = {
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".png": "image/png",
    ".gif": "image/gif",
    ".webp": "image/webp",
    ".bmp": "image/bmp",
    ".tiff": "image/tiff",
}

_ocr_prompt = """
Extract content from image.

Your answer should be a json, containing abstract and full text in such struct: `{"abstract": "xxxx", "full": "xxxx"}`.

The full text should be in markdown syntax. It may has lists, tables, code.

If something looks like list in full text, represent it with legal markdown list syntax.
""".strip()


class Client:
    def __init__(self, name: str, base_url: str, model: str, api_key: str, tools):
        self.name = name
        self.base_url = base_url
        self.model = model
        self.api_key = api_key
        self.tools = tools
        self._client = OpenAI(base_url=base_url, api_key=api_key)

    def stream_response(self, messages: list[dict]):
        stream = self._client.chat.completions.create(
            model=self.model,
            messages=messages,
            tools=self.tools,
            stream=True,
        )

        stream1, stream2 = tee(stream)

        return (
            chunk.choices[0].delta.content
            for chunk in stream1
            if chunk.choices and chunk.choices[0].delta.content
        ), (
            chunk.choices[0].delta.tool_calls
            for chunk in stream2
            if chunk.choices and chunk.choices[0].delta.tool_calls
        )

    def ocr(self, uri, prompt=_ocr_prompt) -> str:
        if os.path.exists(uri):
            with open(uri, "rb") as image_file:
                base64_data = base64.b64encode(image_file.read()).decode("utf-8")
                mime_type = _mime_types.get(Path(uri).suffix.lower(), "image/jpeg")
                uri = f"data:{mime_type};base64,{base64_data}"

        messages = [
            {
                "role": "user",
                "content": [
                    {
                        "type": "image_url",
                        "image_url": {"url": uri},
                    },
                    {"type": "text", "text": prompt},
                ],
            }
        ]

        response = self._client.chat.completions.create(
            model=self.model,
            messages=messages,
            response_format={"type": "json_object"},
        )

        return response.choices[0].message.content



from prompt_toolkit import PromptSession
from prompt_toolkit.styles import Style

style = Style.from_dict(
    {
        "prompt": "ansigreen",  # Blue prompt text
        "": "ansigreen",  # Green input text
    }
)


def session() -> PromptSession:
    return PromptSession(">>> ", style=style)



import inspect
import json
import os
import re
import shutil

from markitdown import MarkItDown
from fastmcp import Client


class LocalTools:
    definitions = [
        {
            "type": "function",
            "function": {
                "name": "read_file",
                "description": "Read data from a local file",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local file",
                        },
                    },
                    "required": ["path"],
                },
            },
        },
        {
            "type": "function",
            "function": {
                "name": "write_file",
                "description": "Write data to local file",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local file",
                        },
                        "data": {
                            "type": "string",
                            "description": "data to write",
                        },
                    },
                    "required": ["path"],
                },
            },
        },
        {
            "type": "function",
            "function": {
                "name": "remove_file",
                "description": "Remove a local file, after user approval",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local file to remove",
                        },
                    },
                    "required": ["path"],
                },
            },
        },
        {
            "type": "function",
            "function": {
                "name": "create_dir",
                "description": "Create a local directory",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local directory to create",
                        },
                    },
                    "required": ["path"],
                },
            },
        },
        {
            "type": "function",
            "function": {
                "name": "read_dir",
                "description": "Read directory structure like tree command",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local directory to read",
                        },
                    },
                    "required": ["path"],
                },
            },
        },
        {
            "type": "function",
            "function": {
                "name": "remove_dir",
                "description": "Remove a local directory, after user approval",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local directory to remove",
                        },
                    },
                    "required": ["path"],
                },
            },
        },
        {
            "type": "function",
            "function": {
                "name": "grep_files",
                "description": "Search for a pattern in a file using regular expressions",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "path to a local file or directory to search in",
                        },
                        "regexp": {
                            "type": "string",
                            "description": "regular expression pattern to search for",
                        },
                    },
                    "required": ["path", "regexp"],
                },
            },
        },
    ]

    @staticmethod
    def read_file(path) -> dict:
        try:
            content = MarkItDown().convert(path).text_content
            return {"status": "success", "data": content, "err": None, "proceed": None}
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}

    @staticmethod
    def write_file(path, data) -> dict:
        try:
            with open(path, "w") as f:
                f.write(data)
            return {"status": "success", "data": None, "err": None, "proceed": None}
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}

    @staticmethod
    def remove_file(path) -> dict:
        try:
            os.remove(path)
            return {"status": "success", "data": None, "err": None, "proceed": None}
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}

    @staticmethod
    def create_dir(path) -> dict:
        try:
            os.makedirs(path, exist_ok=True)
            return {"status": "success", "data": None, "err": None, "proceed": None}
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}

    @staticmethod
    def read_dir(path) -> dict:
        try:
            tree_output = []

            def generate_tree(directory, prefix=""):
                try:
                    items = sorted(os.listdir(directory))
                    items = [item for item in items if not item.startswith(".")]
                except PermissionError:
                    tree_output.append(f"{prefix}[Permission Denied]")
                    return

                for i, item in enumerate(items):
                    item_path = os.path.join(directory, item)
                    is_last = i == len(items) - 1

                    if is_last:
                        tree_output.append(f"{prefix}└── {item}")
                        new_prefix = prefix + "    "
                    else:
                        tree_output.append(f"{prefix}├── {item}")
                        new_prefix = prefix + "│   "

                    if os.path.isdir(item_path):
                        generate_tree(item_path, new_prefix)

            if not os.path.exists(path):
                return {
                    "status": "error",
                    "data": None,
                    "err": "Directory does not exist",
                    "proceed": None,
                }

            if not os.path.isdir(path):
                return {
                    "status": "error",
                    "data": None,
                    "err": "Path is not a directory",
                    "proceed": None,
                }

            tree_output.append(os.path.basename(os.path.abspath(path)) or path)
            generate_tree(path)

            return {
                "status": "success",
                "data": "\n".join(tree_output),
                "err": None,
                "proceed": None,
            }
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}

    @staticmethod
    def remove_dir(path) -> dict:
        try:
            shutil.rmtree(path)
            return {"status": "success", "data": None, "err": None, "proceed": None}
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}

    @staticmethod
    def grep_files(path, regexp) -> dict:
        def get_files(path):
            if os.path.isfile(path):
                yield path
            elif os.path.isdir(path):
                for root, dirs, files in os.walk(path):
                    for file in files:
                        yield os.path.join(root, file)

        def grep_file(path, pattern):
            try:
                with open(path, "r", encoding="utf-8") as f:
                    for line_num, line in enumerate(f, 1):
                        if pattern.search(line):
                            yield f"{file_path}:{line_num}: {line.rstrip()}"
            except (UnicodeDecodeError, PermissionError):
                pass

        try:
            pattern = re.compile(regexp)
            matched_lines = []

            for file_path in get_files(path):
                for line in grep_file(file_path, pattern):
                    matched_lines.append(line)

            return {
                "status": "success",
                "data": matched_lines if matched_lines else "",
                "err": None,
                "proceed": None,
            }
        except Exception as e:
            return {"status": "error", "data": None, "err": str(e), "proceed": None}


class MCPClient:
    def __init__(self, mcp: Client):
        self.client = mcp
        self.specs = None

    async def __aenter__(self):
        await self.client.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client.__aexit__(exc_type, exc_val, exc_tb)

    async def list_tools(self):
        if self.specs is None:
            tools = await self.client.list_tools()
            self.specs = [
                {
                    "type": "function",
                    "function": {
                        "name": t.name,
                        "description": t.description,
                        "parameters": t.inputSchema,
                    },
                }
                for t in tools
            ]
        return self.specs

    async def call_tool(self, name, params):
        result = await self.client.call_tool(name, params, raise_on_error=False)
        if result.is_error:
            err = result.content[0].text
            return {"status": "error", "data": None, "err": err, "proceed": None}

        data = "".join([c.text for c in result.content if c.type == "text"])
        return {"status": "success", "data": data, "err": None, "proceed": None}


class Call:
    def __init__(self, id, fn, params):
        self.id = id
        self.fn = fn
        self.params = params

    def params_str(self) -> str:
        def stringified_value():
            for name, val in self.params.items():
                yield name, str(val)

        return str(
            {
                name: val if len(val) <= 60 else f"{val[:28]}....{val[-28:]}"
                for name, val in stringified_value()
            }
        )


class Calls:
    def __init__(self, stream):
        self.stream = stream
        self.consumed = False
        self.buffer = None

    def __len__(self) -> int:
        return len(self._consume())

    def __iter__(self):
        for id, name, args in self._func_args():
            params = None
            try:
                params = json.loads(args)
            except Exception as e:
                print(args)
                raise e
            yield Call(id, name, params)

    def __str__(self) -> str:
        self._consume()
        return ";".join([f"{name}({args})" for _, name, args in self._func_args()])

    def _func_args(self):
        for fn_call in self._consume().values():
            yield (fn_call["id"], fn_call["fn"]["name"], fn_call["fn"]["args"])

    def _consume(self):
        if self.buffer is None:
            buffer = {}
            for tool in self.stream:
                for t in tool:
                    index = t.index
                    if index not in buffer:
                        buffer[index] = {
                            "id": t.id,
                            "fn": {"name": t.function.name, "args": ""},
                        }
                    if t.function.arguments:
                        buffer[index]["fn"]["args"] += t.function.arguments
            self.buffer = buffer
        return self.buffer


class Tools:
    def __init__(self):
        self.mcps: dict[str, MCPClient] = {}

    async def add_mcp(self, transport):
        if transport in self.mcps.items():
            return
        mcp = MCPClient(Client(transport))
        await mcp.__aenter__()
        self.mcps[transport] = mcp

    async def del_mcp(self, transport):
        mcp = self.mcps.get(transport, None)
        if mcp is None:
            return
        del self.mcps[transport]
        await mcp.__aexit__(None, None, None)

    async def specs(self) -> list:
        defs = [] + LocalTools.definitions
        for _, mcp in self.mcps.items():
            defs += await mcp.list_tools()
        return defs

    async def execute(self, call: Call):
        async def exec(call: Call):
            for _, mcp in self.mcps.items():
                for mcp_tool in await mcp.list_tools():
                    if mcp_tool["function"]["name"] == call.fn:
                        return await mcp.call_tool(call.fn, call.params)

            method = getattr(LocalTools, call.fn)
            valid_params = inspect.signature(method).parameters.keys()
            filtered_params = {
                k: v for k, v in call.params.items() if k in valid_params
            }
            return method(**filtered_params)

        return json.dumps(await exec(call))
